#' @title Maximize a function using ftypes
#'
#' @description This function takes a data.frame (inst) a objective function
#' (lfun), and a method, and returns the value of the method's function,
#' typically the parameters that maximize the function. Most usefully employed
#' to subsequently estimate a covariance matrix.
#'
#' @param inst a data.frame or list with relavant data. Must have elements
#' "ID", and "covars"
#' @param lfun the user defined function to maximize
#' @param init the initial parameters of each ftype
#' @param ftypes the list of function types
#' @param covars a list of ftypes with sets of covariates
#' @param transforms a numeric vector of transforms for the base parameters,
#' again, rcguts style
#' @param method the method used for the optimizer
#' @param optopt options for use with the optimizer
#' @param ... named variables passed directly to the likelihood function
#' @export rcest
rcest <- function(inst,  lfun,
                  init = list(), ftypes = list(), covars = list(),
                  transforms = list(), method, optopt, ...) {

  # Check the instrument
  inst_check(inst)

  # Check the lfun
  req_args <- c("pmat", "inst", "ftypes", "ft_index")
  fn_check(lfun, req_args, "lfun")

  wrap_opts <- get_wrap_opts(inst   = inst,
                             ftypes = ftypes,
                             covars = covars,
                             transforms = transforms,
                             init = init, ...)

  # The pars are just the unlisted sanitized init and covars
  pars <- c(unlist(wrap_opts$made_init), unlist(wrap_opts$made_covars))
  names(pars) <- NULL

  opt <- get_optimizer(method)

  opt_out <- opt$fn(pars = pars, lfun = lfun, wrap_opts = wrap_opts,
                    method = method, optopt = optopt)
  opt_out$ftypes     <- wrap_opts$ftypes
  opt_out$ft_index   <- wrap_opts$ft_index
  opt_out$covars     <- wrap_opts$covars
  opt_out$covars_in  <- wrap_opts$covars_in
  opt_out$cv_index   <- wrap_opts$cv_index
  opt_out$transforms <- wrap_opts$transforms
  opt_out$method     <- method
  opt_out$inst       <- inst

  # Run the function once at the results to get the likelihood vector
  opt_out$like_vec <- lfun_wrap(par = opt_out$estimates, lfun = lfun,
                                wrap_opts = wrap_opts, sum_res = FALSE)

  opt_out
}

#' @title Top level handling of generic likelihood functions
#'
#' This function is meant to be called from another function which does the
#' necessary prep to make sure that the layout is acceptable, and especially
#' that the ftypes of the pars and covars or acceptable.
#'
#' @param par the raw parameter vector
#' @param lfun the user supplied function to maximize
#' @param wrap_opts the wrapped options
#' @param sum_res Does the function need to be summed?
#' @export lfun_wrap
lfun_wrap <- function(par, lfun, wrap_opts, sum_res = TRUE) {

  # Allow wrap opts to be the name of the wrap_opts object
  # Useful for cluster evaluations
  if (! is.list(wrap_opts)) {
    wrap_opts <- get(wrap_opts, envir = .GlobalEnv)
  }

  # Adjust for covariates and transforms
  pmat <- pvec_adjust(par, wrap_opts)
  # Run the user-supplied likelihood funciton
  dargs <- list()
  if (length(wrap_opts$dot_args) > 0) {
    dargs <- wrap_opts$dot_args
  }

  dargs$pmat     <- pmat
  dargs$inst     <- wrap_opts$inst
  dargs$ftypes   <- wrap_opts$ftypes
  dargs$ft_index <- wrap_opts$ft_index

  ll <- do.call(lfun, dargs)
  if (sum_res) {
    ll <- sum(ll)
  }
  ll
}

#' @title Adjust a parameter matrix for covariates
#'
#' @description This will be used almost exclusively in internal functions
#' @param pvec the full parameter vector
#' @param wrap_opts a list generated by get_wrap_opts
#' @export pvec_adjust
pvec_adjust <- function(pvec, wrap_opts) {

  inst       <- wrap_opts$inst
  ft_index   <- wrap_opts$ft_index
  covars     <- wrap_opts$covars
  cv_index   <- wrap_opts$cv_index
  transforms <- wrap_opts$made_transforms

  # Switch major parameters to a matrix
  bpars <- unique(unlist(ft_index))
  bpars <- bpars[bpars > 0]
  pmat <- matrix(pvec[bpars],
                 nrow  = nrow(inst),
                 ncol  = length(bpars),
                 byrow = TRUE)

  pmat_adjust(pmat, pvec, wrap_opts)
}

#' @title Adjust a parameter matrix for covariates
#'
#' @description This will be used almost exclusively in internal functions
#' @param pmat the matrix of base pars
#' @param pvec the full parameter vector
#' @param wrap_opts a list generated by get_wrap_opts
#' @export pmat_adjust
pmat_adjust <- function(pmat, pvec, wrap_opts) {

  inst       <- wrap_opts$inst
  ft_index   <- wrap_opts$ft_index
  covars     <- wrap_opts$covars
  cv_index   <- wrap_opts$cv_index
  transforms <- wrap_opts$made_transforms

  # Adjust the pmat for the covars
  for (ft in names(covars)) {
    # Par index bits for this ftype
    fi  <- ft_index[[ft]]
    # If there are no parameters for this ftype, just continue
    if (fi[1] == 0) next
    # Covar bits for this ftype
    clist <- covars[[ft]]
    cvar_index <- cv_index[[ft]]
    # Transforms for this ftype
    tf <- transforms[[ft]]
    # Go through each parameter for function type ft
    for (i in seq_along(fi)) {
      # The index value for base parater i for function type ft
      bpar_i <- fi[i]
      # The tranform for base parameter i for ftype ft
      ti <- tf[[i]]
      # The index values for the cvars for base par 'i'
      cvar_i <- cvar_index[[i]]
      # The names of the covaritates for the base par
      cn <- names(clist[[i]])
      # Loop through all the covariates, adjust the base par column
      for (i in seq_along(cn)) {
        pmat[, bpar_i] <- pmat[, bpar_i] + pvec[cvar_i[i]] * inst[, cn[i]]
      }
      # Transform this parameter after the covars have been incorporated
      pmat[, bpar_i] <- ti$g(pmat[, bpar_i])
    }
  }

  pmat
}
