library(testthat)
library(rcest)

test_that("get_transform returns the default none and exp transforms", {
  expect_is(get_transform("none"),  "list")
  expect_is(get_transform("exp"),   "list")
  expect_is(get_transform("logit"), "list")

  expect_setequal(names(get_transform("none")),  c("name", "g", "gi", "gp"))
  expect_setequal(names(get_transform("exp")),   c("name", "g", "gi", "gp"))
  expect_setequal(names(get_transform("logit")), c("name", "g", "gi", "gp"))
})

test_that("get_transform returns correct values, and functions are correct for 'none'", {
  none_t <- get_transform("none")
  expect_equal(none_t$name, "none")

  for (i in -10:10) {
    expect_equal(none_t$g(i), i)
    expect_equal(none_t$gi(none_t$g(i)), i)
    expect_equal(none_t$gp(i), 1)
  }
})

test_that("get_transform returns correct values, and functions are correct for 'exp'", {
  none_t <- get_transform("exp")
  expect_equal(none_t$name, "exp")

  for (i in -3:3) {
    expect_equal(none_t$g(i), exp(i))
    expect_equal(none_t$gi(none_t$g(i)), i)
    expect_equal(none_t$gp(i), exp(i))
  }
})

test_that("get_transform returns correct values, and functions are correct for 'logit'", {
  none_t <- get_transform("logit")
  expect_equal(none_t$name, "logit")

  for (i in -3:3) {
    expect_equal(none_t$g(i), 1 / (1 + exp(-i)))
    expect_equal(none_t$gi(none_t$g(i)), i)
    expect_equal(none_t$gp(i), exp(-i) / (exp(-i) + 1)^2)
  }
})

test_that("correctly defined transforms set, get, and return correct values", {
  expect_silent(set_transform(name = "test_1",
                              g    = function(x) 2 * x + 1,
                              gi   = function(x) (x - 1) / 2,
                              gp   = function(x) 2))

  test_1t <- get_transform("test_1")
  expect_equal(test_1t$name, "test_1")

  for (i in -3:3) {
    expect_equal(test_1t$g(i), 2 * i + 1)
    expect_equal(test_1t$gi(test_1t$g(i)), i)
    expect_equal(test_1t$gp(i), 2)
  }
})

test_that("correctly defined transforms set, get, and return correct values", {

  expect_error(set_transform(name = 1234,
                              g    = function(x) 2 * x + 1,
                              gi   = function(x) (x - 1) / 2,
                              gp   = function(x) 2),
  "The 'name' argument of the transform must be a string")

  expect_error(set_transform(name = "test_1",
                              g    = "test",
                              gi   = function(x) (x - 1) / 2,
                              gp   = function(x) 2),
  "The 'g' argument of the transform must be a function")

  expect_error(set_transform(name = "test_1",
                              g    = function(x) 2 * x + 1,
                              gi   = "test",
                              gp   = function(x) 2),
  "The 'gi' argument of the transform must be a function")

  expect_error(set_transform(name = "test_1",
                              g    = function(x) 2 * x + 1,
                              gi   = function(x) (x - 1) / 2,
                              gp   = "none"),
  "The 'gp' argument of the transform must be a function or NULL")
})

test_that("A NULL gp will return the derivitive of 'g'", {
  set_and_get <- function(i) {
    set_transform(name = "test_1",
                  g    = function(x) x^3,
                  gi   = function(x) x ^ (1 / 3),
                  gp   = NULL)
    tt1 <- get_transform(name = "test_1")
    tt1$gp(i)
  }
  for (i in -3:3) {
    expect_equal(set_and_get(i), 3 * i^2)
  }
})

test_that("avail_transform gives the names of the transforoms available", {
  trans <- c("exp", "none", "logit", "test_1")
  expect_setequal(trans, avail_transform())
})
