library(testthat)
library(rcest)

###### Parser setting is checked in previous test #####
## Clear the initial environments
set_rcenv()

set_parser(ftype = "test_fun1",
           fname = c("fun12", "FUN12"),
           transforms = c("exp", "none"),
           def_pars = c(1, 0),
           par_names = c("p1", "p2"),
           description = "This is a test function with 2 parameters")

set_parser(ftype = "test_fun2",
           fname = c("fun21", "FUN21"),
           transforms = c("exp", "none"),
           def_pars = c(1, 2),
           par_names = c("p1", "par2"),
           description = "This is a second test function with 2 parameters")

set_parser(ftype = "test_fun3",
           fname = c("fun31", "FUN31"),
           transforms = c("none"),
           def_pars = c(1),
           par_names = c("p1"),
           description = "This is a third test function with 1 parameter")

set_parser(ftype = "test_fun4",
           fname = c("fun41", "FUN41"),
           transforms = rep("none", 3),
           def_pars = seq_len(3),
           par_names = paste0("p", seq_len(3)),
           description = "This is a fourth test function with 3 parameters")

set_parser(ftype = "test_fun5",
           fname = c("fun51", "FUN51"),
           transforms = rep("none", 2),
           def_pars = seq_len(2),
           par_names = paste0("q", seq_len(2)),
           description = "This is a fifth test function, one that is meant to not have covars")


##### Check covars functions #####

# covars are ftype like lists. The names of the toplevel elements are the names
# of ftypes. The elements themselves are either lists where each element
# corresponds to a parameter of the ftype (named and order-agnostic or unnamed
# and order-dependent), or a vector of covariates to be applied to every
# parameter in the ftype.

# Some internals
test_that("vectors of cvars are turned into named numeric vectors", {

  # Check character vectors are initialized at 0
  cvar     <- c("dem1", "dem2")
  expected <- c("dem1" = 0, "dem2" = 0)
  have     <- rcest:::cvar_vec_to_numeric(cvar)
  expect_equal(expected, have)

  # Check numeric vectors are just passed through
  cvar     <- c("dem1" = -1, "dem2" = .3)
  expected <- c("dem1" = -1, "dem2" = .3)
  have     <- rcest:::cvar_vec_to_numeric(cvar)
  expect_equal(expected, have)

  # Check mixed vectors have non-initialized elements set to 0
  # And initialized elements correctly set
  cvar     <- c("dem1" = 1, "dem2",     "dem3" = 2)
  expected <- c("dem1" = 1, "dem2" = 0, "dem3" = 2)
  have     <- rcest:::cvar_vec_to_numeric(cvar)
  expect_equal(expected, have)

  # Check empty and null vectors return null vectors
  # And initialized elements correctly set
  cvar     <- c()
  expected <- c()
  have     <- rcest:::cvar_vec_to_numeric(cvar)
  expect_equal(expected, have)

  cvar     <- vector(mode = "numeric", length = 0)
  expected <- c()
  have     <- rcest:::cvar_vec_to_numeric(cvar)
  expect_equal(expected, have)
})

test_that("vector cvars are turned into lists", {

  ftypes   <- list("test_fun1" = "fun12",
                   "test_fun3" = "fun31")

  covars   <- list("test_fun1" = c("dem1",     "dem2"),
                   "test_fun3" = c("dem1" = 2, "dem2" = 1))

  # First test that a character vec cvar is turned into a list
  fname    <- "test_fun1"
  expected <- list("p1" = c("dem1" = 0, "dem2" = 0),
                   "p2" = c("dem1" = 0, "dem2" = 0))
  have     <- rcest:::cvar_vec_to_list(covars[[fname]], fname, ftypes)
  expect_equal(expected, have)

  # Next test that a numeric vec cvar is turned into a list
  fname    <- "test_fun3"
  expected <- list("p1"   = c("dem1" = 2, "dem2" = 1))
  have     <- rcest:::cvar_vec_to_list(covars[[fname]], fname, ftypes)
  expect_equal(expected, have)

})

ftypes   <- list("test_fun1" = "fun12",
                  "test_fun2" = "fun21",
                  "test_fun3" = "fun31",
                  "test_fun4" = "fun41")

test_that("a fully named list is filled out and expanded", {
  covars   <- list("test_fun1" = list("p1" = c("dem1", "dem2"),
                                      "p2" = c("dem3", "dem4")))

  ftype    <- "test_fun1"
  expected <- list("p1" = c("dem1" = 0, "dem2" = 0),
                   "p2" = c("dem3" = 0, "dem4" = 0))
  have <- rcest:::cvar_list_to_list(covars[[ftype]], ftype, ftypes)
  expect_equal(expected, have)
})

test_that("a partially named list is filled out and expanded", {
  covars   <- list("test_fun4" = list("p1" = c("dem1", "dem2"),
                                             c("dem3", "dem4"),
                                             c("dem5", "dem6")
                                             ))

  ftype    <- "test_fun4"
  expected <- list("p1" = c("dem1" = 0, "dem2" = 0),
                   "p2" = c("dem3" = 0, "dem4" = 0),
                   "p3" = c("dem5" = 0, "dem6" = 0))
  have <- rcest:::cvar_list_to_list(covars[[ftype]], ftype, ftypes)

  expect_equal(expected, have)
})

test_that("list with fewer parameters than expected is filled out", {
  covars   <- list("test_fun4" = list(c("dem1")))

  ftype    <- "test_fun4"
  expected <- list("p1" = c("dem1" = 0),
                   "p2" = c(),
                   "p3" = c())
  have <- rcest:::cvar_list_to_list(covars[[ftype]], ftype, ftypes)

  expect_equal(expected, have)
})

test_that("an initialized list is filled out and expanded", {
  covars   <- list("test_fun4" = list(c("dem1" = 1.1, "dem2" = 2.2),
                                      c(),
                                      c("dem5" = 5.5, "dem6" = 6.6)
                                      ))

  ftype    <- "test_fun4"
  expected <- list("p1" = c("dem1" = 1.1, "dem2" = 2.2),
                   "p2" = c(),
                   "p3" = c("dem5" = 5.5, "dem6" = 6.6))
  have <- rcest:::cvar_list_to_list(covars[[ftype]], ftype, ftypes)

  expect_equal(expected, have)
})

test_that("an out-of-order initialized list is filled out and expanded", {
  covars   <- list("test_fun4" = list(c(),
                                      c("dem5" = 5.5, "dem6" = 6.6),
                                      "p1" = c("dem1" = 1.1, "dem2" = 2.2)
                                      ))

  ftype    <- "test_fun4"
  expected <- list("p1" = c("dem1" = 1.1, "dem2" = 2.2),
                   "p2" = c(),
                   "p3" = c("dem5" = 5.5, "dem6" = 6.6))
  have <- rcest:::cvar_list_to_list(covars[[ftype]], ftype, ftypes)

  expect_equal(expected, have)
})

# covar_sanitize
test_that("correctly specified covar_sanitize returns sanitized covariates for estimation.", {
    # While complex, this is a supported covar initialization
    covars   <- list("test_fun1" = "dem1",
                     "test_fun2" = list(c("dem1" = 3)),
                     "test_fun3" = c("dem1" = 1, "dem2" = 2),
                     "test_fun4" = list("p3" = c("dem5" = 5.5, "dem6" = 6.6),
                                        c("dem1" = 1.1, "dem2" = 2.2),
                                        c()))

    # While complex, this is a supported covar initialization
    expected <- list("test_fun1" = list("p1" = c("dem1" = 0),  "p2"   = c("dem1" = 0)),
                     "test_fun2" = list("p1" = c("dem1" = 3),  "par2" = c()),
                     "test_fun3" = list("p1" = c("dem1" = 1,   "dem2" = 2)),
                     "test_fun4" = list("p1" = c("dem1" = 1.1, "dem2" = 2.2),
                                        "p2" = c(),
                                        "p3" = c("dem5" = 5.5, "dem6" = 6.6)),
                     "test_fun5" = list("q1" = c(),
                                        "q2" = c())
    )

    ftypes <- list(test_fun1 = "fun12",
                   test_fun2 = "fun21",
                   test_fun3 = "fun31",
                   test_fun4 = "fun41",
                   test_fun5 = "fun51")

    have <- covar_sanitize(covars = covars, ftypes = ftypes)
    # Only "test_fun1" is in the covars
    expect_equal(have, expected)
})

test_that("covar_sanitize fails when covars isn't a list", {
    # Check that non-lists fail
    covars <- data.frame()
    expect_error(covar_sanitize(covars), "'covars' needs to be a ftype list")
    covars <- vector("numeric", 12)
    expect_error(covar_sanitize(covars), "'covars' needs to be a ftype list")
    covars <- c()
    expect_error(covar_sanitize(covars), "'covars' needs to be a ftype list")
})

test_that("covar_sanitize fails an element of covars is unnamed", {
    covars <- list(test_fun1 = "dem1",  "dem2")
    # Check that non-lists fail
    msg <- "Every element in 'covars' must be named for an ftype"
    expect_error(covar_sanitize(covars, ftypes), msg)
})

test_that("covar_sanitize fails an ftype is in covars and not in ftypes", {
    ftypes <- list(test_fun1 = "fun12", test_fun2 = "fun21")
    covars <- list(test_fun1 = "dem1",  test_fun0 = "dem2")
    # Check that non-lists fail
    msg <- "ftypes 'test_fun0' are in covars but not in ftypes"
    expect_error(covar_sanitize(covars, ftypes), msg)
})

test_that("covar_sanitize fails when a covar isn't a character/numeric vector or a list", {
    ftypes <- list(test_fun1 = "fun12", test_fun2 = "fun21")
    covars <- list(test_fun1 = c(1),       test_fun2 = "dem2")
    # Check that non-lists fail
    msg <- paste0("The covar for ftype 'test_fun1' is not a list,",
                  " or named numeric vector or character vector")
    expect_error(covar_sanitize(covars, ftypes), msg)

    covars <- list(test_fun1 = T,       test_fun2 = "dem2")
    # Check that non-lists fail
    msg <- paste0("The covar for ftype 'test_fun1' is not a list,",
                  " or named numeric vector or character vector")
    expect_error(covar_sanitize(covars, ftypes), msg)
})

test_that("covar_sanitize fails when a covar is a list but it's elements are proper vectors", {
    ftypes <- list(test_fun1 = "fun12", test_fun2 = "fun21")
    covars <- list(test_fun1 = list(p1 = T), test_fun2 = "dem2")
    # Check that elements are the right types
    msg <- paste0("Element 1 for covar ftype 'test_fun1' is not NULL, ",
                  "a numeric or character vector, or a cvb or cbvec")
    expect_error(covar_sanitize(covars, ftypes), msg)
})

# covar_index
test_that("covar_index returns right index with correctly specified covars", {
    ftypes <- list(test_fun1 = "fun12", test_fun2 = "fun21")
    covars <- list(test_fun1 = list(p1 = c("dem1", "dem2")),
                   test_fun2 = "dem2")

    expect <- list(test_fun1 = list(p1 = c(5, 6), p2 = c()),
                   test_fun2 = list(p1 = c(7), par2 = c(8))
    )

    have <- covar_index(covars, ftypes)
    expect_equal(have, expect)

    ftypes <- list(test_fun3 = "fun31")
    covars <- list(test_fun3 = "dem2")
    expect <- list(test_fun3 = list(p1 = 2))
    have <- covar_index(covars, ftypes)
    expect_equal(have, expect)

})
