library(testthat)
library(rcest)

test_that("the gradient of a function is properly calculated with a single-parameter function", {
  x <- c(1)

  # First derivitive of exp is exp, so easy to test
  f <- function(x) exp(x)
  g <- grad(f, x)
  expect_equal(g, f(x))

  # First derivitive of a linear function is just the coefficient
  f <- function(x) 2 + 3 * x
  g <- grad(f, x)
  expect_equal(g, 3)

  # First derivitive of a quadratic function is a little more complicated
  f <- function(x) 2 + 3 * x ^2
  g <- grad(f, x)
  expect_equal(g, 6 * x)

})

test_that("the gradient of a function is properly calculated with multi-parameter function", {
  # Multi-parameter function
  x <- c(1, 2)

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]
  g <- grad(f, x)
  expect_equal(g, c(6, 2))

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]^3
  g <- grad(f, x)
  expect_equal(g, c(6, 24))

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]^3 + 4 * x[1]^4 * x[2]^2
  g <- grad(f, x)
  expect_equal(g, c(70, 40))
})

test_that("the grad function fails when fn doeesn't return scalar numeric output", {
  # Two parameter output
  x <- c(1, 2)
  f <- function(x) x
  expect_error(grad(f, x),
  "The function passed to 'grad' does not return a scalar value. Maybe you need 'jacobian'?")

  x <- c(1, 2)
  f <- function(x) NA
  expect_error(grad(f, x),
               "The function passed to 'grad' does not return a numeric value at the inital parameters")

})

test_that("the grad function fails when fn doeesn't return numeric output", {
  # Two parameter output
  x <- c(1, 2)
  f <- function(x) "hello world!"
  expect_error(grad(f, x),
               "The function passed to 'grad' does not return a numeric value at the inital parameters")

  x <- c(1, 2)
  f <- function(x) NA
  expect_error(grad(f, x),
               "The function passed to 'grad' does not return a numeric value at the inital parameters")

})

test_that("the jacobian of a multi-input/output function is properly calculated", {
  x <- c(3, 7)

  f <- function(x) {
    y <- vector("numeric", 3)
    y[1] <- 1 + x[1] ^2 + 1 * x[2]^3
    y[2] <- 1 + x[1] ^3 + 2 * x[2]^3
    y[3] <- 1 + x[1] ^4 + 3 * x[2]^3
    y
  }
  j <- jacobian(f, x)
  j0 <- matrix(c(1 * 2 * x[1],   1 * 3 * x[2]^2,
                 1 * 3 * x[1]^2, 2 * 3 * x[2]^2,
                 1 * 4 * x[1]^3, 3 * 3 * x[2]^2),
               byrow = T, nrow = 3)
  expect_equal(j, j0)

  # Make some random input
  x <- rbinom(2, 10, .5)
  f <- function(x) {
    y <- vector("numeric", 3)
    y[1] <- 1 + x[1] ^2 + 1 * x[2]^3
    y[2] <- 1 + x[1] ^3 + 2 * x[2]^3
    y[3] <- 1 + x[1] ^4 + 3 * x[2]^3
    y
  }
  j <- jacobian(f, x)
  j0 <- matrix(c(1 * 2 * x[1],   1 * 3 * x[2]^2,
                 1 * 3 * x[1]^2, 2 * 3 * x[2]^2,
                 1 * 4 * x[1]^3, 3 * 3 * x[2]^2),
               byrow = T, nrow = 3)
  expect_equal(j, j0)

})


test_that("the hessian of a function is properly calculated", {
  x <- c(1)

  # First derivitive of exp is exp, so easy to test
  f <- function(x) exp(x)
  h <- hessian(f, x)
  expect_equal(h, f(x))

  # First derivitive of a linear function is just the coefficient
  f <- function(x) 2 + 3 * x
  h <- hessian(f, x)
  expect_equal(h, 0)

  # First derivitive of a quadratic function is a little more complicated
  f <- function(x) 2 + 3 * x ^2
  h <- hessian(f, x)
  expect_equal(h, 6)

  # Multi-parameter function
  x <- c(1, 2)

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]
  h <- hessian(f, x)
  h0 <- matrix(c(6, 0,
                 0, 0),
               byrow = T, nrow = 2)
  expect_equal(h, h0)

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]^3
  h <- hessian(f, x)
  h0 <- matrix(c(6, 0,
                 0, 12 * x[2]),
               byrow = T, nrow = 2)
  expect_equal(h, h0)

  f <- function(x) 2 + 3 * x[1]^2 + 2 * x[2]^3 + 4 * x[1]^4 * x[2]^2
  h <- hessian(f, x)
  h0 <- matrix(c(48 * x[1]^2 * x[2]^2 + 6, 32 * x[1]^3 * x[2],
                 32 * x[1]^3 * x[2],       8 * x[1]^4 + 12 * x[2]),
               byrow = T, nrow = 2)
  expect_equal(h, h0)
})
