library(testthat)
library(rcest)

# Basic tests for setting the cvb
test_that("covar builder sets a proper function in the environment", {
  fn <- function(inst, val) {
    vars <- inst[, 1:3]
    init <- rep(0, 3)
    list(vars = var, init = init)
  }
  expect_invisible(set_cvb(c("Test1", "t1"), fn))
})

test_that("the names are set correctly", {
  fn <- function(inst, val) {
    vars <- inst[, 1:3]
    init <- rep(0, 3)
    list(vars = var, init = init)
  }
  expect_error(set_cvb(c(1:2), fn), "'names' must be a character vector")
})

test_that("a function is passed when setting", {
  expect_error(set_cvb(c("Test1", "t1"), 0), "'fn' must be a function")
})

test_that("improper argument list throws an error", {
  fn <- function(val) {
    val
  }
  expect_error(set_cvb(c("Test1", "t1"), fn), "'fn' is missing arguments: 'inst'")
})

# Checking cvb is set
test_that("the cvb is listed as available", {
  # Not a proper cvb function, but won't error out until called
  fn <- function(inst, val) {
    list(vars = 1, init = 1)
  }
  set_cvb(c("Test1", "t1"), fn)
  defaults <- c("identity", "id",
                "dummy", "d",
                "polynomial", "poly", "p",
                "interaction", "inter", "x")
  avail <- c(defaults, "Test1", "t1")
  expect_setequal(avail_cvb(), avail)
})

test_that("the cvb constructer returns a cvb class object", {
  expect_invisible(v <- cvb("id", "dem1"))
  expect_equivalent(class(v), "cvb")
  expect_setequal(slotNames(v), c("name", "dargs"))
})

test_that("the cvb constructer fails when the cvb hasn't been set", {
  emsg <- "cvb 'fail' not in covar environment. Did you forget to set it?"
  expect_error(v <- cvb("fail"), emsg)
})

test_that("concatenating multiple cvb returns a proper cbvec", {
  cv1 <- cvb("id", "dem1")
  cv2 <- cvb("id", "dem2")
  v <- c(cv1, cv2)
  # Check the class
  expect_equivalent(class(v), "cbvec")
  expect_setequal(slotNames(v), "values")
  # Check all the values are now cvbs
  values <- slot(v, "values")
  for (v in values) {
    expect_equivalent(class(v), "cvb")
    expect_setequal(slotNames(v), c("name", "dargs"))
  }
})

test_that("concatenating cvb with regular elements returns a proper cbvec", {
  cv1 <- cvb("id", "dem1")
  cv2 <- cvb("id", "dem2")
  expect_invisible(v <- c(cv1, cv2, "dem3", "dem4" = 1))
  # Check the class
  expect_equivalent(class(v), "cbvec")
  expect_setequal(slotNames(v), "values")
  # Check all the values are now cvbs
  values <- slot(v, "values")
  for (v in values) {
    expect_equivalent(class(v), "cvb")
    expect_setequal(slotNames(v), c("name", "dargs"))
  }

  # Check that the non-cvb's were properly transformed
  expect_setequal(slot(values[[3]], "name"), "identity")
  expect_setequal(slot(values[[4]], "name"), "identity")

  expect_equal(slot(values[[3]], "dargs"), list("dem3" = 0))
  expect_equal(slot(values[[4]], "dargs"), list("dem4" = 1))
})

test_that("cvb cannot be combined with non-character/numeric elements", {
  cv1 <- cvb("id", "dem1")
  cv2 <- cvb("id", "dem2")
  expect_error(c(cv1, cv2, T))
  expect_error(c(cv1, cv2, list()))
  expect_error(c(cv1, cv2, data.frame(one = 1)))
})

test_that("covar_sanitize works with cvb", {
  set_parser(
    ftype = "test",
    fname = "t1",
    transforms = c("none", "none"),
    def_pars = c(0, 1),
    par_names = c("t1p1", "t1p2"),
    description = "Test function 1"
  )

  set_parser(
    ftype = "test",
    fname = "t2",
    transforms = c("none", "none", "none"),
    def_pars = c(0, 1, 2),
    par_names = c("t2p1", "t2p2", "t2p3"),
    description = "Test function 2"
  )

  ftypes <- list(test = "t1")
  covars <- list(test = list(
    t1p1 = c(
      cvb("d", "dem1"),
      cvb("p", "dem2", 2),
      "dem2"
    ),
    t1p2 = c()
  ))

  # Test if c combines multiple items correctly and sanitizes correctly
  cbtest <- list(cvb("d", "dem1"), cvb("p", "dem2", 2), cvb("identity", "dem2" = 0))
  cbtest <- new("cbvec", values = cbtest)

  covars <- covar_sanitize(covars, ftypes = ftypes)
  cv_expect <- list(test = list(t1p1 = cbtest, t1p2 = NULL))
  expect_equal(covars, cv_expect)

  # Test if cvb can be used on its own
  covars <- list(test = cvb("d", "dem1"))

  cbtest2 <- list(cvb("d", "dem1"))
  cbtest2 <- new("cbvec", values = cbtest2)

  covars    <- covar_sanitize(covars, ftypes = ftypes)
  cv_expect <- list(test = list(t1p1 = cbtest2, t1p2 = cbtest2))

  expect_equal(covars, cv_expect)

  # Test if cvb can be used on its own in a list of pars
  covars <- list(test = list(
    t1p1 = c(
      cvb("d", "dem1"),
      cvb("p", "dem2", 2),
      "dem2"
    ),
    t1p2 = cvb("d", "dem1")
  ))

  cbtest2 <- list(cvb("d", "dem1"))
  cbtest2 <- new("cbvec", values = cbtest2)

  covars    <- covar_sanitize(covars, ftypes = ftypes)
  cv_expect <- list(test = list(t1p1 = cbtest, t1p2 = cbtest2))

  expect_equal(covars, cv_expect)

})

test_that("correct covars are adusted correctly", {
  set_parser(
    ftype = "test",
    fname = "t1",
    transforms = c("none", "none"),
    def_pars = c(0, 1),
    par_names = c("p1", "p2"),
    description = "Test function"
  )

  ftypes <- list(test = "t1")
  covars <- list(test = list(
    p1 = c(
      cvb("d", "dem1"),
      cvb("p", "dem2", 2),
      "dem2"
    ),
    p2 = c()
  ))

  covars_adj <- list(test = list(
    p1 = c(
      cvb("dummy", var = "dem1", include = 1:2),
      cvb("p", "dem2", 2),
      "dem2"
    ),
    p2 = c()
  ))

  covars <- covar_sanitize(covars, ftypes = ftypes)
  covars_adj <- covar_sanitize(covars_adj, ftypes = ftypes)

  inst <- data.frame(
    dem1 = c(1, 1, 1, 0, 0, 0, 0, 0, 2, 2, 2, 2),
    dem2 = c(1, 1, 1, 2, 2, 2, 3, 3, 3, 0, 0, 0)
  )

  iout    <- inst_adjust(covars, inst)
  ninst   <- iout$inst
  ncovars <- iout$covars_o

  # Check that the instrument names are as expected
  expect_setequal(names(ninst), c("dem1", "dem2", "dem1@1", "dem1@2", "dem2^2"))
  # Check that the out covars are as expected
  expect_setequal(names(ncovars$test$p1), c("dem1@1", "dem1@2", "dem2^2", "dem2"))
  # Check the in covars is returned adjusted
  expect_equal(covars_adj, iout$covars_i)

  # Check that dummy returns the right variables
  expect_setequal(ninst[["dem1@1"]], ifelse(inst[["dem1"]] == 1, 1, 0))
  expect_setequal(ninst[["dem1@2"]], ifelse(inst[["dem1"]] == 2, 1, 0))
  # Check that poly returns the right variables
  expect_setequal(ninst[["dem2^2"]], inst[["dem2"]]^2)
})
