library(testthat)
library(rcest)

# ensure the default environment is set
set_rcdefault()

# Some simulated data. The equation that generated this is
# y ~ dem1 + dem2 + rnorm(n, 0, s)

inst <- data.frame(ID = c(1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4,
                          4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 8,
                          8, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10),
                   dem1 = c(-0.42, -0.42, -0.42, -0.42, -0.42, 0.87, 0.87,
                            0.87, 0.87, 0.87, -0.04, -0.04, -0.04, -0.04,
                            -0.04, -1.58, -1.58, -1.58, -1.58, -1.58, -0.18,
                            -0.18, -0.18, -0.18, -0.18, -1.46, -1.46, -1.46,
                            -1.46, -1.46, 0.73, 0.73, 0.73, 0.73, 0.73, -1.25,
                            -1.25, -1.25, -1.25, -1.25, 0.33, 0.33, 0.33, 0.33,
                            0.33, 0.2, 0.2, 0.2, 0.2, 0.2),
                   dem2 = c(1.12, 1.12, 1.12, 1.12, 1.12, -0.19, -0.19, -0.19,
                            -0.19, -0.19, -0.51, -0.51, -0.51, -0.51, -0.51,
                            0.34, 0.34, 0.34, 0.34, 0.34, 1.14, 1.14, 1.14,
                            1.14, 1.14, -0.22, -0.22, -0.22, -0.22, -0.22,
                            0.37, 0.37, 0.37, 0.37, 0.37, 2.27, 2.27, 2.27,
                            2.27, 2.27, 1.06, 1.06, 1.06, 1.06, 1.06, 1.87,
                            1.87, 1.87, 1.87, 1.87),
                   s = c(1.94, 1.94, 1.94, 1.94, 1.94, 2.02, 2.02, 2.02, 2.02,
                         2.02, 1.97, 1.97, 1.97, 1.97, 1.97, 2.04, 2.04, 2.04,
                         2.04, 2.04, 2.11, 2.11, 2.11, 2.11, 2.11, 1.96, 1.96,
                         1.96, 1.96, 1.96, 2, 2, 2, 2, 2, 1.97, 1.97, 1.97,
                         1.97, 1.97, 1.79, 1.79, 1.79, 1.79, 1.79, 2.08, 2.08,
                         2.08, 2.08, 2.08),
                   y = c(0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0,
                         1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1,
                         0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1))

# Set the parser for a logistic regression
set_parser(ftype = "logit",
           fname = "logit",
           transforms = c("none"),
           def_pars = 0,
           par_names = "y0",
           description = "Logit regression",
           set_default = F)

# Define the likelihood function
logit_like <- function(pmat, inst, ftypes, ft_index) {
  yhat <- 1 / (1 + exp(-pmat[, ft_index$logit]))
  log(ifelse(inst$y == 1, yhat, 1 - yhat))
}

test_that("rcest returns the same results as glm for a logit", {
  # Estimate the likelihood
  ftypes <- list(logit = "logit")
  init   <- list(logit = c(0.5))
  covars <- list(logit = c("dem1" = .5, "dem2" = -.25))
  # Optimizer Options
  optopt <- list(report = F, iterations = 10000)
  method <- "NR"
  # Do the estimation
  res    <- rcest(inst = inst, lfun = logit_like,
                  init = init, ftypes = ftypes,
                  covars = covars, method = method,
                  optopt = optopt)
  # Recover the covariance matrix, without clustering
  covmat <- calc_covmat(res, lfun = logit_like, inst = inst)
  # Make the results presentable
  parsed <- make_parsed(res, covmat)

  mod <- glm(y ~ dem1 + dem2, data = inst, family = binomial(link = "logit"))

  est_glm <- mod$coefficients
  est_res <- res$estimates

  expect_true(all(round(est_res, 4) - round(est_glm, 4) == 0))

  cov_glm <- matrix(summary(mod)$cov.scaled, nrow = 3, ncol = 3)
  cov_res <- covmat

  expect_equal(cov_res, cov_glm, 9e-7)

  se_glm <- sqrt(diag(summary(mod)$cov.scaled))
  se_res <- parsed$SE
  names(se_glm) <- NULL
  names(se_res) <- NULL

  expect_equal(se_res, se_glm, 9e-7)

})

test_that("rcest returns correct clustered standard errors", {
  # Estimate the likelihood
  ftypes <- list(logit = "logit")
  init   <- list(logit = c(0.5))
  covars <- list(logit = c("dem1" = .5, "dem2" = -.25))
  # Optimizer Options
  optopt <- list(report = F, iterations = 10000)
  method <- "NR"
  # Do the estimation
  res    <- rcest(inst = inst, lfun = logit_like,
                  init = init, ftypes = ftypes,
                  covars = covars, method = method,
                  optopt = optopt)
  # Recover the covariance matrix, with clustering
  covmat <- calc_covmat(res, lfun = logit_like, inst = inst, cluster = "ID")
  # Make the results presentable
  parsed <- make_parsed(res, covmat)
  parsed$SE

  se_cluster <- c(0.4050708, 0.3947014, 0.4135215)
  expect_equal(parsed$SE, se_cluster, tolerance = 9e-7)
})
